﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace CANSimModbusDemo
{
    public partial class MainForm : Form
    {
        private EasyModbus.ModbusClient modbusClient = null;
        CANSim.Knob speedKnob = new CANSim.Knob(CANSim.Knob.SpeedMaxValue);
        CANSim.Knob rpmKnob = new CANSim.Knob(CANSim.Knob.RpmMaxValue);
        CANSim.Knob backlightKnob = new CANSim.Knob(CANSim.Knob.BacklightMaxValue);
        CANSim.Knob steeringKnob = new CANSim.Knob(CANSim.Knob.SteeringMaxValue);
        CANSim.Knob p5Knob = new CANSim.Knob(CANSim.Knob.P5MaxValue);
        CANSim.Knob p6Knob = new CANSim.Knob(CANSim.Knob.P6MaxValue);
        private bool protocolLogChecked = true;
        private bool autoScrollChecked = true;

        public MainForm()
        {
            InitializeComponent();
            modbusClient = new EasyModbus.ModbusClient();
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            comboDir.SelectedIndex = 0;
            radioB10.Checked = true;
            radioB20.Checked = true;
            radioB30.Checked = true;
            radioB40.Checked = true;
            controlEnable(false);
        }

        private void DisconnectClient()
        {
            timerRead.Enabled = false;
            if (comboDir.SelectedIndex != 0) //reading
            {
                // enable HW controls
                modbusClient.WriteSingleRegister((int)CANSim.RegiestersAdresses.HoldingRegisters.DeviceConfiguration, 0);
            }
            modbusClient.Disconnect();
        }


        private void buttonConnect_Click(object sender, EventArgs e)
        {
            // connection and disconnetion CANSim device
            try
            {
                if (modbusClient.Connected)
                {
                    DisconnectClient();
                    buttonConnect.Text = "Connect";
                    logMessage("COM port closed");
                }
                else
                {
                    modbusClient.SerialPort = "COM" + numericUpDownCOM.Value;
                    modbusClient.Parity = CANSim.SerialPortParameters.parity;
                    modbusClient.StopBits = CANSim.SerialPortParameters.stopBits;
                    modbusClient.UnitIdentifier = CANSim.SerialPortParameters.modbusAddress;
                    modbusClient.Baudrate = CANSim.SerialPortParameters.baudrate;
                    modbusClient.Connect();
                    buttonConnect.Text = "Disconnect";
                    logMessage("COM port opened");
                    directionChanged();
                }

            }
            catch (Exception ex)
            {
                logMessage(ex.GetType() + ": " + ex.Message);
            }
        }

        private void writeKnobValue(KnobControl.KnobControl knobControl)
        {
            int value = 0;
            int addr = 0;
            if (knobControl == knobSpeed)
            {
                addr = (int)CANSim.RegiestersAdresses.HoldingRegisters.SpeedKnobPosition;
                value = speedKnob.Value2Knob(knobControl.Value);
            }
            else if (knobControl == knobRpm)
            {
                addr = (int)CANSim.RegiestersAdresses.HoldingRegisters.RpmKnobPosition;
                value = rpmKnob.Value2Knob(knobControl.Value * 100);
            }
            else if (knobControl == knobBacklight)
            {
                addr = (int)CANSim.RegiestersAdresses.HoldingRegisters.BacklightKnobPosition;
                value = backlightKnob.Value2Knob(knobControl.Value);
            }
            else if (knobControl == knobSteering)
            {
                addr = (int)CANSim.RegiestersAdresses.HoldingRegisters.SteeringKnobPosition;
                value = steeringKnob.Value2Knob(knobControl.Value);
            }
            else if (knobControl == knobP5)
            {
                addr = (int)CANSim.RegiestersAdresses.HoldingRegisters.P5KnobPosition;
                value = p5Knob.Value2Knob(knobControl.Value);
            }
            else if (knobControl == knobP6)
            {
                addr = (int)CANSim.RegiestersAdresses.HoldingRegisters.P6KnobPosition;
                value = p6Knob.Value2Knob(knobControl.Value);
            }
            logMessage("", "Knob changing to value: " + knobControl.Value + " (writing register on address " + addr + ", value " + value + ").");
            modbusClient.WriteSingleRegister(addr, value);
        }

        private void knob_ValueChanged(object Sender)
        {
            if (!timerRead.Enabled && modbusClient.Connected)
            {
                writeKnobValue((KnobControl.KnobControl) Sender);
            }
        }

        private void timerRead_Tick(object sender, EventArgs e)
        {
            //reads values from ipnut registers
            if (comboDir.SelectedIndex == 0 && modbusClient.Connected)
            {
                // read speed knob
                int[] inputRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.SpeedKnobPosition, 1);
                logMessage("", "Reading knob value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.SpeedKnobPosition + ").");
                if (inputRegister.Length > 0) knobSpeed.Value = speedKnob.Knob2Value(inputRegister[0]);
                // read rpm knob
                inputRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.RpmKnobPosition, 1);
                logMessage("", "Reading knob value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.RpmKnobPosition + ").");
                if (inputRegister.Length > 0) knobRpm.Value = rpmKnob.Knob2Value(inputRegister[0]) / 100;
                // read backlight knob
                inputRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.BacklightKnobPosition, 1);
                logMessage("", "Reading knob value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.BacklightKnobPosition + ").");
                if (inputRegister.Length > 0) knobBacklight.Value = backlightKnob.Knob2Value(inputRegister[0]);
                // read Steering knob
                inputRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.SteeringKnobPosition, 1);
                logMessage("", "Reading knob value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.SteeringKnobPosition + ").");
                if (inputRegister.Length > 0) knobSteering.Value = steeringKnob.Knob2Value(inputRegister[0]);
                // read P5 knob
                inputRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.P5KnobPosition, 1);
                logMessage("", "Reading knob value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.P5KnobPosition + ").");
                if (inputRegister.Length > 0) knobP5.Value = p5Knob.Knob2Value(inputRegister[0]);
                // read P6 knob
                inputRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.P6KnobPosition, 1);
                logMessage("", "Reading knob value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.P6KnobPosition + ").");
                if (inputRegister.Length > 0) knobP6.Value = p6Knob.Knob2Value(inputRegister[0]);
                // read switches bitmap
                int[] switchesRegister = modbusClient.ReadInputRegisters((int)CANSim.RegiestersAdresses.InputRegisters.SwitchesPosition, 1);
                logMessage("", "Reading switches value (reading input register on address " + (int)CANSim.RegiestersAdresses.InputRegisters.SwitchesPosition + ").");
                if (((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosKL15) != 0) && ((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosKLS) != 0))
                {
                    radioB12.Checked = true;
                    radioB11.Checked = false;
                    radioB10.Checked = false;
                }
                else if ((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosKLS) != 0)
                {
                    radioB12.Checked = false;
                    radioB11.Checked = true;
                    radioB10.Checked = false;
                }
                else
                {
                    radioB12.Checked = false;
                    radioB11.Checked = false;
                    radioB10.Checked = true;
                }
                if ((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosSleep) != 0)
                {
                    radioB21.Checked = true;
                    radioB20.Checked = false;
                }
                else
                {
                    radioB21.Checked = false;
                    radioB20.Checked = true;
                }
                if ((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosBacklight) != 0)
                {
                    radioB31.Checked = true;
                    radioB30.Checked = false;
                }
                else
                {
                    radioB31.Checked = false;
                    radioB30.Checked = true;
                }
                if ((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosPDC) != 0)
                {
                    radioB42.Checked = true;
                    radioB41.Checked = false;
                    radioB40.Checked = false;
                }
                else if ((switchesRegister[0] & 1 << CANSim.BitPositions.bitPosREV) != 0)
                {
                    radioB42.Checked = false;
                    radioB41.Checked = true;
                    radioB40.Checked = false;
                }
                else
                {
                    radioB42.Checked = false;
                    radioB41.Checked = false;
                    radioB40.Checked = true;
                }
            }
        }

        private void comboDir_SelectedValueChanged(object sender, EventArgs e)
        {
            if (modbusClient.Connected)
            {
                directionChanged();
            }
        }

        private void controlEnable(bool enable)
        {
            knobSpeed.Enabled = enable;
            knobRpm.Enabled = enable;
            knobBacklight.Enabled = enable;
            knobSteering.Enabled = enable;
            knobP5.Enabled = enable;
            knobP6.Enabled = enable;
            radioB10.AutoCheck = enable;
            radioB11.AutoCheck = enable;
            radioB12.AutoCheck = enable;
            radioB20.AutoCheck = enable;
            radioB21.AutoCheck = enable;
            radioB30.AutoCheck = enable;
            radioB31.AutoCheck = enable;
            radioB40.AutoCheck = enable;
            radioB41.AutoCheck = enable;
            radioB42.AutoCheck = enable;
        }

        void logMessage(string msg, string protocolMsg = "")
        {
            if (protocolLogChecked)
            {
                msg += " " + protocolMsg;
            }
            if (msg.Length > 0)
            {
                listBox.Items.Add(msg);
                if (autoScrollChecked) listBox.TopIndex = listBox.Items.Count - 1;
            }
        }
        void directionChanged()
        {
            if (comboDir.SelectedIndex == 0) //reading
            {
                logMessage("CANSim HW control enabling", "(writing register on address " + (int)CANSim.RegiestersAdresses.HoldingRegisters.DeviceConfiguration +", value 0).");
                modbusClient.WriteSingleRegister((int)CANSim.RegiestersAdresses.HoldingRegisters.DeviceConfiguration, 0);
                logMessage("Starting reading timer.");
                timerRead.Enabled = true;
                controlEnable(false);
            }
            else //writing
            {
                logMessage("Stoping reading timer.");
                timerRead.Enabled = false;
                logMessage("CANSim HW control disabling", "(writing register on address " + (int)CANSim.RegiestersAdresses.HoldingRegisters.DeviceConfiguration + ", value 1).");
                modbusClient.WriteSingleRegister((int)CANSim.RegiestersAdresses.HoldingRegisters.DeviceConfiguration, 1);
                controlEnable(true);
                writeKnobValue(knobSpeed);
                writeKnobValue(knobRpm);
                writeKnobValue(knobBacklight);
                writeKnobValue(knobSteering);
                writeKnobValue(knobP5);
                writeKnobValue(knobP6);
                writeSwitch();
            }
        }

        void writeSwitch()
        {
            int value;
            value = (radioB12.Checked) ? 1 << CANSim.BitPositions.bitPosKL15 | 1 << CANSim.BitPositions.bitPosKLS : 0;  // KL15, KLS
            value |= (radioB11.Checked) ? 1 << CANSim.BitPositions.bitPosKLS : 0; // KLS
            value |= (radioB21.Checked) ? 1 << CANSim.BitPositions.bitPosSleep : 0; // Sleep
            value |= (radioB31.Checked) ? 1 << CANSim.BitPositions.bitPosBacklight : 0; // Backlight
            value |= (radioB42.Checked) ? 1 << CANSim.BitPositions.bitPosPDC : 0; // PDC
            value |= (radioB41.Checked) ? 1 << CANSim.BitPositions.bitPosREV : 0; // REV
            logMessage("State of switches changing", "(writing register on address " + (int)CANSim.RegiestersAdresses.HoldingRegisters.SwitchesPosition + ", value " + value + ").");
            modbusClient.WriteSingleRegister((int)CANSim.RegiestersAdresses.HoldingRegisters.SwitchesPosition, value);
        }

        private void radio_CheckedChanged(object sender, EventArgs e)
        {
            if (!timerRead.Enabled && modbusClient.Connected)
            {
                RadioButton radioButton = (RadioButton)sender;
                if (radioButton.Checked)
                    writeSwitch();
            }
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if(modbusClient.Connected)
            {
                DisconnectClient();
            }
        }

        private void checkBoxProtocolLog_CheckedChanged(object sender, EventArgs e)
        {
            protocolLogChecked = checkBoxProtocolLog.Checked;
        }

        private void checkBoxAutoScroll_CheckedChanged(object sender, EventArgs e)
        {
            autoScrollChecked = checkBoxAutoScroll.Checked;
        }
    }
}